package controller;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;

import model.Language;
import model.MorphoTags;
import model.StateAnnotation;
import model.WordNet;
import model.ontology.Ontology;
import nlputil.Tokenizer;
import nlputil.morphotagger.Mxpost;
import nlputil.morphotagger.TreeTagger;

import org.annolab.tt4j.TreeTaggerException;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import view.AnnotersView;
import view.LimitAnnotationView;
import view.TextView;
import view.components.Text;
import view.components.Tokens;
import edu.smu.tspell.wordnet.Synset;

/**
 * Class to control the principal functions of software
 * @author gonta
 *
 */
public class MainController {

	private boolean open;
	
	private JFrame frame;
	private JFileChooser chooser;
	
	private FileFilter xml_filter;
	private FileFilter txt_filter;
	
	/**POS-Taggers for english and spanish */	
	private TreeTagger english_tagger;
	private TreeTagger spanish_tagger;
	private TreeTagger portuguese_tagger;
	
	/**necessary files for the pos-taggers*/
	private File mxpost_tag_set;

	/**
	 * 	Constructor 
	 * @param frame 
	 * @throws IOException
	 */
	public MainController( JFrame frame ) throws IOException{
		
		this.open = false;
		this.frame = frame;
		this.chooser = new JFileChooser();
		
		
		//Creating the pos-taggers for portuguese, english and spanish
		this.mxpost_tag_set = new File("lib"+File.separator+"MXPOST"+File.separator+"port");
		File english_model = new File( "lib" +File.separator+ "treetagger" + File.separator + "model" +  File.separator + "english.par:utf8");
		File spanish_model = new File(  "lib" +File.separator+ "treetagger" + File.separator + "model" +  File.separator + "spanish.par:utf8");
		File portuguese_model = new File(  "lib" +File.separator+ "treetagger" + File.separator + "model" +  File.separator + "pt.par:utf8"); 
		
		this.english_tagger = new TreeTagger( english_model );
		this.spanish_tagger = new TreeTagger( spanish_model );				
		this.portuguese_tagger = new TreeTagger( portuguese_model );
		
		/* Filter of xml files*/
		this.xml_filter = new FileFilter() {			
			@Override
			public String getDescription() {
				return "xml";
			}
			@Override
			public boolean accept(File f) {				
				if ( f.isDirectory() )
					return true;
				
				String name_spit[] = f.getName().split("\\.");
				if ( name_spit.length >= 2 )
					return ( name_spit[ 1 ].compareToIgnoreCase("xml") == 0 );
				
				return false;
			}
		};
		
		/* Filter of txt files*/
		this.txt_filter = new FileFilter() {			
			@Override
			public String getDescription() {
				return "txt";
			}
			@Override
			public boolean accept(File f) {				
				if ( f.isDirectory() )
					return true;
				
				String name_spit[] = f.getName().split("\\.");
				if ( name_spit.length >= 2 )
					return ( name_spit[ 1 ].compareToIgnoreCase("txt") == 0 ) ;
				
				return false;
			}
		};
		
	}
	
	
	/**
	 * Method to load a file or a fileset that will be annotated.
	 * files loaded will be tokenized, tagged with mxpost or tagger and after this process, will be loaded for 
	 * the annotation
	 * 
	 * @param textView UI to visualize the texts
	 * @throws IllegalArgumentException if some selection mistake occur
	 */
	public void openFile( TextView textView, AnnotersView annotersView, LimitAnnotationView limitAnnotationView ) throws IllegalArgumentException{
				
		if ( chooser.isShowing() ) chooser.setVisible(false);
		
		chooser.setDialogTitle("Escolhe um ou mais textos");
		chooser.setFileFilter( txt_filter );
		chooser.setMultiSelectionEnabled( true );
		int option = chooser.showOpenDialog( frame );
				
		if ( option != JFileChooser.APPROVE_OPTION )
			return;
		
		File file[] = chooser.getSelectedFiles();
		if ( file == null )
			throw new IllegalArgumentException("No file selected!");

		if ( file.length == 0 )
			throw new IllegalArgumentException("No file selected!");
		
		try {
			
			File out_dir = new File( file[ 0 ].getAbsoluteFile().getParentFile().getAbsolutePath() + File.separator+ "annotation");
			File annotationFile = new File( file[ 0 ].getAbsoluteFile().getParentFile().getAbsolutePath() + File.separator+ "annotation" + File.separator + "annotation.xml");

			if( annotationFile.exists() ) {				
			
				JOptionPane.showMessageDialog( frame, "Existe uma anotação prévia, procederá a carregar-se", "Anotação Existente", JOptionPane.INFORMATION_MESSAGE);
				openAnnotation(textView, annotersView, limitAnnotationView , true);
				
			} else {
				if ( !out_dir.exists() )
					out_dir.mkdir();
				
				//Vetor onde será armazenados os arquivos anotados com o mxpost
				File pos_file[] = new File[ file.length ];
							
				//For every file, is executed the annotation with the mxpost file
				for( int i = 0; i < file.length; i++ ){

					//tokenize the file
					File input_tokenizada = new File("tokens_temp.txt");
					Tokenizer.tokenizer( file[ i ], input_tokenizada );
					System.out.println(input_tokenizada.getAbsolutePath());

					String name_file = file[ i ].getName().split("\\.")[ 0 ];
					Language language_file = Language.convert( name_file.split("_")[ 0 ]);
					if (language_file == null ) language_file = Language.PORTUGUESE;

					pos_file[ i ] = new File( out_dir.getAbsolutePath() + File.separator + name_file + ".tagged");
					
					switch ( language_file) {

						case PORTUGUESE:
							Mxpost.tagger_file( input_tokenizada, pos_file[ i ] , mxpost_tag_set );
						break;

						case ENGLISH:
							english_tagger.taggerFile( input_tokenizada, pos_file[ i ] );
						break;
						
						case SPANISH:
							spanish_tagger.taggerFile( input_tokenizada, pos_file[ i ] );
						break;
					}
					
				}
				
				textView.setText( pos_file );
				
				this.open = true;
			}
			
		} catch (IOException e) {
						
			String erro = "";
			for( int i=0; i < file.length; i++)
				erro+= "<li>" + file[ i ].getName() + "</li>";

			
			System.out.println("Erro:\n" + erro + "\n\n" + e.getMessage() );
			
			throw new IllegalArgumentException("<html>Error in open file:<ul>" + erro + "</ul>");
			
		} catch (TreeTaggerException e) {		
			e.printStackTrace();
		} catch (JDOMException e) {
			e.printStackTrace();
		}
		
		frame.paintAll( frame.getGraphics() );
	}


	/**
	 * This method starts the process o openning of a annotation saved previouslly.
	 * In this method, use a filter that allows just the openning of xml files
	 * 
	 * @param text_view Component when will be inserted the texts
	 * @param annoters_view Component when will be inserted the annotators
	 * 
	 * @throws IllegalArgumentException
	 * @throws IOException
	 * @throws JDOMException
	 * @throws IOException
	 */
	public void openAnnotation( TextView text_view, AnnotersView annoters_view, LimitAnnotationView limitAnnotation_view, boolean auto ) throws IllegalArgumentException, IOException, JDOMException, IOException{
		
		File [] files;
		File annotationFile = null;
		if ( !auto ) {		
			if ( chooser.isShowing() ) chooser.setVisible(false);;
			
			chooser.setDialogTitle("Select a xml file");
			chooser.setFileFilter( xml_filter );
			chooser.setMultiSelectionEnabled( false );
			chooser.showOpenDialog( frame );
			files = new File [1];
			files[0] = chooser.getSelectedFile();
			
			if ( files == null )
				throw new IllegalArgumentException("File not found. Please, for open a annoted file, you have select a xml file.");
			
			if ( files.length == 1 )
				annotationFile = files[0];
		} else {
			files = chooser.getSelectedFiles();						
			annotationFile = new File( files[ 0 ].getAbsoluteFile().getParentFile().getAbsolutePath() + File.separator+ "annotation" + File.separator + "annotation.xml");
		}
		

		text_view.enabledAnnotateTypes();
		//Handling the xml file for loading the annotation
		SAXBuilder builder = new SAXBuilder();
		Document document = builder.build( new FileInputStream( annotationFile ) );

		Element root = document.getRootElement();

		List<Element> list = root.getChild("Anotadores").getChildren("Anotador");		
		for( int i=0; i < list.size(); i++ )
			annoters_view.addAnnotator( list.get( i ).getText() );
		
		list = root.getChild("LimitesAnotacao").getChildren("Pos");		
		for( int i=0; i < list.size(); i++ )
			limitAnnotation_view.setLimits( list.get(i).getAttributeValue("id"), Double.parseDouble( list.get(i).getText() ) );
	
		list = root.getChild("Arquivos").getChildren("Texto");
		
		Text array_text[] = new Text[ list.size() ];
		
		int index = 0;
		String translate = null;
		String translate_set[] = null;
		Synset synset = null;
		LinkedList<Synset> synset_set = null;
		Tokens t;
		boolean manual_translate;
		for( Element text : list ){
			
			array_text[ index ] = new Text(
					text.getAttributeValue("name") , 
					annotationFile.getParentFile() );
			
			Language text_language = Language.valueOf( text.getAttributeValue("language") );

			for( Element paragraph : (List<Element>) text.getChildren("p") ){
				
				StateAnnotation annoted_state = StateAnnotation.NO_ANNOTATED;
				for( Element token : (List<Element>) paragraph.getChildren("Token") ){
					
					annoted_state= StateAnnotation.valueOf( token.getChildText("Type") );
					
					t = new Tokens(
							token.getChildText("Valor"),
							annoted_state,
							token.getChildText("Tag"),
							MorphoTags.strToTag( token.getChildText("MorphoTag") ),
							text_language
							);
					
					t.setPartOfComplexPredicate(token.getChild("Valor").getAttributeValue("cp"));
					
					t.setLemma( token.getChildText("Lema") );
					
					String obsComment = token.getChild( "Comment" ).getAttributeValue("obs").trim();
					String contentComment = token.getChild( "Comment" ).getAttributeValue("content").trim();
					
					t.setComment(model.Comment.getNameTag(contentComment) );
					t.setCommentObs( obsComment );
					
					t.setMorphoTagPOS(MorphoTags.strToTag( token.getChildText("MorphoTagPOS") ));
					
					Element translate_set_element = token.getChild("Traducoes");
					if( translate_set_element.getAttributeValue("traducao_manual").compareTo("true") == 0)
						manual_translate = false;
					else
						manual_translate = true;
					
					List<Element> translate_element = translate_set_element.getChildren();
					if( !translate_element.isEmpty() ){
						
						translate_set = new String[ translate_element.size() ];
						for( int i=0; i < translate_element.size(); i++ ){
							
							if( translate_element.get(i).getAttributeValue("selecionado").compareTo("true")  == 0)
								translate = translate_element.get( i ).getText();							
						
							translate_set[ i ] = translate_element.get( i ).getText();
						}
						
						Element synset_set_element = token.getChild("Synsets");
						List<Element> synset_element = synset_set_element.getChildren();
						synset_set = new LinkedList<Synset>();
						try{
						for( int i=0; i < synset_element.size(); i++ ){
							
							System.out.println(">"+ synset_element.get( i ).getText() +"<");
							
							if( synset_element.get(i).getAttributeValue("selecionado").compareTo("true")  == 0)
								synset = WordNet.getSynset( Integer.parseInt( synset_element.get( i ).getText() ) , t.getMorphoTag().name() );
							
							synset_set.add( WordNet.getSynset( Integer.parseInt( synset_element.get( i ).getText() ) , t.getMorphoTag().name() ) );
						}
						}catch(Exception e ){
							System.out.println(e.getMessage());
						}
						t.setAnnotation( manual_translate, translate, synset, translate_set, synset_set );
						t.setType( annoted_state );
					}
					
					array_text[ index ].addToken( t );
				}
				
				array_text[ index ].addToken( null );
			}
			
			index++;
		}
		
		this.open = true;
		Ontology g = Ontology.getInstance();		
		File origem = new File(annotationFile.getParent() + File.separator + annotationFile.getName().replace(".xml","_ontology_all.xml"));
		g.load(origem);
		
		text_view.setText( array_text );
		
	}

	/**
	 * 
	 * Save the annotation of loaded files
	 * The files saved will have the same name of source file, but will be added 
	 * a word ANOTADO at the end of file.
	 * @param text_view component that contains the annotation
	 */
	public void save( TextView text_view ) {
		
		for( Text t : text_view.getTexts() ){
			try {
				
				File out = t.getDirectory();
				
				if( !out.exists() )
					out.mkdir();
				
				out = new File( 
						t.getDirectory().getAbsolutePath() +
						File.separator +			 
						t.getNameFile().split("\\.")[0] + "_ANOTADO");
				
				BufferedWriter writer = new BufferedWriter(
						new OutputStreamWriter( new FileOutputStream( out ), "UTF-8" ));

				for( Tokens token : t.getTokens() ){
					if ( token == null )
						writer.write("\n\n");
					else
						writer.write( token.write() +" ");
				}
				
				writer.close();

			} catch (UnsupportedEncodingException e) {			
				e.printStackTrace();
				
			} catch (FileNotFoundException e) {			
				e.printStackTrace();
				
			} catch (IOException e) {			
				e.printStackTrace();
			}
		}
		
		JOptionPane.showMessageDialog( frame, "Anotação salva com sucesso!", "Anotação", JOptionPane.INFORMATION_MESSAGE);
		
	}

	public boolean isOpen() {
		return this.open;
		
	}
}
